@@ -5,6 +5,7 @@ import (
 	"bytes"
 	"errors"
 	"fmt"
+	"io/fs"
 	"os"
 	"os/exec"
 	"runtime"
@@ -13,7 +14,7 @@ import (
 	"text/template"
 
 	"github.com/coreos/go-semver/semver"
-	vfs "github.com/twpayne/go-vfs/v2"
+	vfs "github.com/twpayne/go-vfs/v3"
 	"go.uber.org/multierr"
 )
 
@@ -23,7 +24,7 @@ type SourceState struct {
 	system                  System
 	sourceDirAbsPath        AbsPath
 	destDirAbsPath          AbsPath
-	umask                   os.FileMode
+	umask                   fs.FileMode
 	encryption              Encryption
 	ignore                  *patternSet
 	minVersion              semver.Version
@@ -95,6 +96,8 @@ func WithTemplateOptions(templateOptions []string) SourceStateOption {
 	}
 }
 
+// A targetStateEntryFunc returns a TargetStateEntry based on reading an AbsPath
+// on a System.
 type targetStateEntryFunc func(System, AbsPath) (TargetStateEntry, error)
 
 // NewSourceState creates a new source state with the given options.
@@ -119,25 +122,25 @@ type PreAddFunc func(targetRelPath RelPath, newSourceStateEntry, oldSourceStateE
 
 // AddOptions are options to SourceState.Add.
 type AddOptions struct {
-	AutoTemplate     bool
-	Create           bool
-	Empty            bool
-	Encrypt          bool
-	EncryptedSuffix  string
-	Exact            bool
-	Include          *EntryTypeSet
-	PreAddFunc       PreAddFunc
-	RemoveDir        RelPath
-	Template         bool
-	TemplateSymlinks bool
+	AutoTemplate     bool          // Automatically create templates, if possible.
+	Create           bool          // Add create_ entries instead of normal entries.
+	Empty            bool          // Add the empty_ attribute to added files.
+	Encrypt          bool          // Encrypt files.
+	EncryptedSuffix  string        // Suffix for encrypted files.
+	Exact            bool          // Add the exact_ attribute to added directories.
+	Include          *EntryTypeSet // Only add types in this set.
+	PreAddFunc       PreAddFunc    // Function to be called before the source entry is added.
+	RemoveDir        RelPath       // Directory to remove before adding.
+	Template         bool          // Add the .tmpl attribute to added files.
+	TemplateSymlinks bool          // Add symlinks with targets in the source or home directories as templates.
 }
 
 // Add adds destAbsPathInfos to s.
-func (s *SourceState) Add(sourceSystem System, persistentState PersistentState, destSystem System, destAbsPathInfos map[AbsPath]os.FileInfo, options *AddOptions) error {
+func (s *SourceState) Add(sourceSystem System, persistentState PersistentState, destSystem System, destAbsPathInfos map[AbsPath]fs.FileInfo, options *AddOptions) error {
 	type update struct {
 		destAbsPath    AbsPath
 		entryState     *EntryState
-		sourceRelPaths SourceRelPaths
+		sourceRelPaths []SourceRelPath
 	}
 
 	destAbsPaths := make([]AbsPath, 0, len(destAbsPathInfos))
@@ -192,7 +195,7 @@ DESTABSPATH:
 		update := update{
 			destAbsPath:    destAbsPath,
 			entryState:     entryState,
-			sourceRelPaths: SourceRelPaths{sourceEntryRelPath},
+			sourceRelPaths: []SourceRelPath{sourceEntryRelPath},
 		}
 
 		if oldSourceStateEntry, ok := s.entries[targetRelPath]; ok {
@@ -280,9 +283,9 @@ DESTABSPATH:
 	return nil
 }
 
-// AddDestAbsPathInfos adds an os.FileInfo to destAbsPathInfos for destAbsPath
+// AddDestAbsPathInfos adds an fs.FileInfo to destAbsPathInfos for destAbsPath
 // and any of its parents which are not already known.
-func (s *SourceState) AddDestAbsPathInfos(destAbsPathInfos map[AbsPath]os.FileInfo, system System, destAbsPath AbsPath, info os.FileInfo) error {
+func (s *SourceState) AddDestAbsPathInfos(destAbsPathInfos map[AbsPath]fs.FileInfo, system System, destAbsPath AbsPath, info fs.FileInfo) error {
 	for {
 		if _, err := destAbsPath.TrimDirPrefix(s.destDirAbsPath); err != nil {
 			return err
@@ -322,7 +325,7 @@ type PreApplyFunc func(targetRelPath RelPath, targetEntryState, lastWrittenEntry
 type ApplyOptions struct {
 	Include      *EntryTypeSet
 	PreApplyFunc PreApplyFunc
-	Umask        os.FileMode
+	Umask        fs.FileMode
 }
 
 // Apply updates targetRelPath in targetDir in destSystem to match s.
@@ -425,7 +428,7 @@ func (s *SourceState) Apply(targetSystem, destSystem System, persistentState Per
 					if actualStateDir.perm.Perm() == targetStateDir.perm.Perm() {
 						if targetStateDir.perm.Perm() != lastWrittenEntryState.Mode.Perm() {
 							if targetStateDir.perm.Perm() == lastWrittenEntryState.Mode.Perm()&^s.umask {
-								lastWrittenEntryState.Mode = os.ModeDir | targetStateDir.perm
+								lastWrittenEntryState.Mode = fs.ModeDir | targetStateDir.perm
 								_ = persistentStateSet(persistentState, EntryStateBucket, []byte(targetAbsPath), lastWrittenEntryState)
 							}
 						}
@@ -486,14 +489,23 @@ func (s *SourceState) ExecuteTemplateData(name string, data []byte) ([]byte, err
 	if err != nil {
 		return nil, err
 	}
+
 	for name, t := range s.templates {
 		tmpl, err = tmpl.AddParseTree(name, t.Tree)
 		if err != nil {
 			return nil, err
 		}
 	}
+
+	// Temporarily set .chezmoi.sourceFile to the name of the template.
+	templateData := s.TemplateData()
+	if chezmoiTemplateData, ok := templateData["chezmoi"].(map[string]interface{}); ok {
+		chezmoiTemplateData["sourceFile"] = name
+		defer delete(chezmoiTemplateData, "sourceFile")
+	}
+
 	var sb strings.Builder
-	if err = tmpl.ExecuteTemplate(&sb, name, s.TemplateData()); err != nil {
+	if err = tmpl.ExecuteTemplate(&sb, name, templateData); err != nil {
 		return nil, err
 	}
 	return []byte(sb.String()), nil
@@ -523,7 +535,7 @@ func (s *SourceState) MustEntry(targetRelPath RelPath) SourceStateEntry {
 func (s *SourceState) Read() error {
 	info, err := s.system.Lstat(s.sourceDirAbsPath)
 	switch {
-	case os.IsNotExist(err):
+	case errors.Is(err, fs.ErrNotExist):
 		return nil
 	case err != nil:
 		return err
@@ -533,7 +545,7 @@ func (s *SourceState) Read() error {
 
 	// Read all source entries.
 	allSourceStateEntries := make(map[RelPath][]SourceStateEntry)
-	if err := Walk(s.system, s.sourceDirAbsPath, func(sourceAbsPath AbsPath, info os.FileInfo, err error) error {
+	if err := Walk(s.system, s.sourceDirAbsPath, func(sourceAbsPath AbsPath, info fs.FileInfo, err error) error {
 		if err != nil {
 			return err
 		}
@@ -547,7 +559,7 @@ func (s *SourceState) Read() error {
 
 		parentSourceRelPath, sourceName := sourceRelPath.Split()
 		// Follow symlinks in the source directory.
-		if info.Mode()&os.ModeType == os.ModeSymlink {
+		if info.Mode().Type() == fs.ModeSymlink {
 			// Some programs (notably emacs) use invalid symlinks as lockfiles.
 			// To avoid following them and getting an ENOENT error, check first
 			// if this is an entry that we will ignore anyway.
@@ -678,7 +690,7 @@ func (s *SourceState) Read() error {
 					targetRelPath: destEntryRelPath,
 				})
 			}
-		case os.IsNotExist(err):
+		case errors.Is(err, fs.ErrNotExist):
 			// Do nothing.
 		default:
 			return err
@@ -699,11 +711,13 @@ func (s *SourceState) Read() error {
 		if len(sourceStateEntries) == 1 {
 			continue
 		}
-		sourceRelPaths := make(SourceRelPaths, 0, len(sourceStateEntries))
+		sourceRelPaths := make([]SourceRelPath, 0, len(sourceStateEntries))
 		for _, sourceStateEntry := range sourceStateEntries {
 			sourceRelPaths = append(sourceRelPaths, sourceStateEntry.SourceRelPath())
 		}
-		sort.Sort(sourceRelPaths)
+		sort.Slice(sourceRelPaths, func(i, j int) bool {
+			return sourceRelPaths[i].relPath < sourceRelPaths[j].relPath
+		})
 		err = multierr.Append(err, &errDuplicateTarget{
 			targetRelPath:  targetRelPath,
 			sourceRelPaths: sourceRelPaths,
@@ -814,7 +828,7 @@ func (s *SourceState) addTemplateData(sourceAbsPath AbsPath) error {
 
 // addTemplatesDir adds all templates in templateDir to s.
 func (s *SourceState) addTemplatesDir(templatesDirAbsPath AbsPath) error {
-	return Walk(s.system, templatesDirAbsPath, func(templateAbsPath AbsPath, info os.FileInfo, err error) error {
+	return Walk(s.system, templatesDirAbsPath, func(templateAbsPath AbsPath, info fs.FileInfo, err error) error {
 		if err != nil {
 			return err
 		}
@@ -864,7 +878,7 @@ func (s *SourceState) addVersionFile(sourceAbsPath AbsPath) error {
 	return nil
 }
 
-// applyAll updates targetDir in fs to match s.
+// applyAll updates targetDir in targetSystem to match s.
 func (s *SourceState) applyAll(targetSystem, destSystem System, persistentState PersistentState, targetDir AbsPath, options ApplyOptions) error {
 	for _, targetRelPath := range s.TargetRelPaths() {
 		switch err := s.Apply(targetSystem, destSystem, persistentState, targetDir, targetRelPath, options); {
@@ -898,12 +912,15 @@ func (s *SourceState) newSourceStateDir(sourceRelPath SourceRelPath, dirAttr Dir
 	}
 }
 
+// newCreateTargetStateEntryFunc returns a targetStateEntryFunc that returns a
+// file with sourceLazyContents if the file does not already exist, or returns
+// the actual file's contents unchanged if the file already exists.
 func (s *SourceState) newCreateTargetStateEntryFunc(fileAttr FileAttr, sourceLazyContents *lazyContents) targetStateEntryFunc {
 	return func(destSystem System, destAbsPath AbsPath) (TargetStateEntry, error) {
 		contents, err := destSystem.ReadFile(destAbsPath)
 		switch {
 		case err == nil:
-		case os.IsNotExist(err):
+		case errors.Is(err, fs.ErrNotExist):
 			contents, err = sourceLazyContents.Contents()
 			if err != nil {
 				return nil, err
@@ -919,6 +936,8 @@ func (s *SourceState) newCreateTargetStateEntryFunc(fileAttr FileAttr, sourceLaz
 	}
 }
 
+// newFileTargetStateEntryFunc returns a targetStateEntryFunc that returns a
+// file with sourceLazyContents.
 func (s *SourceState) newFileTargetStateEntryFunc(sourceRelPath SourceRelPath, fileAttr FileAttr, sourceLazyContents *lazyContents) targetStateEntryFunc {
 	return func(destSystem System, destAbsPath AbsPath) (TargetStateEntry, error) {
 		contentsFunc := func() ([]byte, error) {
@@ -942,13 +961,15 @@ func (s *SourceState) newFileTargetStateEntryFunc(sourceRelPath SourceRelPath, f
 	}
 }
 
+// newModifyTargetStateEntryFunc returns a targetStateEntryFunc that returns a
+// file with the contents modified by running the sourceLazyContents script.
 func (s *SourceState) newModifyTargetStateEntryFunc(sourceRelPath SourceRelPath, fileAttr FileAttr, sourceLazyContents *lazyContents) targetStateEntryFunc {
 	return func(destSystem System, destAbsPath AbsPath) (TargetStateEntry, error) {
 		contentsFunc := func() (contents []byte, err error) {
 			// Read the current contents of the target.
 			var currentContents []byte
 			currentContents, err = destSystem.ReadFile(destAbsPath)
-			if err != nil && !os.IsNotExist(err) {
+			if err != nil && !errors.Is(err, fs.ErrNotExist) {
 				return
 			}
 
@@ -1004,6 +1025,8 @@ func (s *SourceState) newModifyTargetStateEntryFunc(sourceRelPath SourceRelPath,
 	}
 }
 
+// newScriptTargetStateEntryFunc returns a targetStateEntryFunc that returns a
+// script with sourceLazyContents.
 func (s *SourceState) newScriptTargetStateEntryFunc(sourceRelPath SourceRelPath, fileAttr FileAttr, targetRelPath RelPath, sourceLazyContents *lazyContents) targetStateEntryFunc {
 	return func(destSystem System, destAbsPath AbsPath) (TargetStateEntry, error) {
 		contentsFunc := func() ([]byte, error) {
@@ -1027,6 +1050,8 @@ func (s *SourceState) newScriptTargetStateEntryFunc(sourceRelPath SourceRelPath,
 	}
 }
 
+// newSymlinkTargetStateEntryFunc returns a targetStateEntryFunc that returns a
+// symlink with the linkname sourceLazyContents.
 func (s *SourceState) newSymlinkTargetStateEntryFunc(sourceRelPath SourceRelPath, fileAttr FileAttr, sourceLazyContents *lazyContents) targetStateEntryFunc {
 	return func(destSystem System, destAbsPath AbsPath) (TargetStateEntry, error) {
 		linknameFunc := func() (string, error) {
@@ -1088,7 +1113,12 @@ func (s *SourceState) newSourceStateFile(sourceRelPath SourceRelPath, fileAttr F
 	}
 }
 
-func (s *SourceState) newSourceStateDirEntry(info os.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
+// newSourceStateDirEntry returns a SourceStateEntry constructed from a
+// directory in s.
+//
+// We return a SourceStateEntry rather than a *SourceStateDir to simplify nil
+// checks later.
+func (s *SourceState) newSourceStateDirEntry(info fs.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
 	dirAttr := DirAttr{
 		TargetName: info.Name(),
 		Exact:      options.Exact,
@@ -1104,14 +1134,19 @@ func (s *SourceState) newSourceStateDirEntry(info os.FileInfo, parentSourceRelPa
 	}, nil
 }
 
-func (s *SourceState) newSourceStateFileEntryFromFile(actualStateFile *ActualStateFile, info os.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
+// newSourceStateFileEntryFromFile returns a SourceStateEntry constructed from a
+// file in s.
+//
+// We return a SourceStateEntry rather than a *SourceStateFile to simplify nil
+// checks later.
+func (s *SourceState) newSourceStateFileEntryFromFile(actualStateFile *ActualStateFile, info fs.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
 	fileAttr := FileAttr{
 		TargetName: info.Name(),
 		Empty:      options.Empty,
 		Encrypted:  options.Encrypt,
 		Executable: isExecutable(info),
 		Private:    isPrivate(info),
-		Template:   options.Template || options.AutoTemplate,
+		Template:   options.Template,
 	}
 	if options.Create {
 		fileAttr.Type = SourceFileTypeCreate
@@ -1123,7 +1158,11 @@ func (s *SourceState) newSourceStateFileEntryFromFile(actualStateFile *ActualSta
 		return nil, err
 	}
 	if options.AutoTemplate {
-		contents = autoTemplate(contents, s.TemplateData())
+		var replacements bool
+		contents, replacements = autoTemplate(contents, s.TemplateData())
+		if replacements {
+			fileAttr.Template = true
+		}
 	}
 	if len(contents) == 0 && !options.Empty {
 		return nil, nil
@@ -1148,7 +1187,12 @@ func (s *SourceState) newSourceStateFileEntryFromFile(actualStateFile *ActualSta
 	}, nil
 }
 
-func (s *SourceState) newSourceStateFileEntryFromSymlink(actualStateSymlink *ActualStateSymlink, info os.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
+// newSourceStateFileEntryFromSymlink returns a SourceStateEntry constructed
+// from a symlink in s.
+//
+// We return a SourceStateEntry rather than a *SourceStateFile to simplify nil
+// checks later.
+func (s *SourceState) newSourceStateFileEntryFromSymlink(actualStateSymlink *ActualStateSymlink, info fs.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
 	linkname, err := actualStateSymlink.Linkname()
 	if err != nil {
 		return nil, err
@@ -1157,8 +1201,7 @@ func (s *SourceState) newSourceStateFileEntryFromSymlink(actualStateSymlink *Act
 	template := false
 	switch {
 	case options.AutoTemplate:
-		contents = autoTemplate(contents, s.TemplateData())
-		fallthrough
+		contents, template = autoTemplate(contents, s.TemplateData())
 	case options.Template:
 		template = true
 	case !options.Template && options.TemplateSymlinks:
@@ -1191,7 +1234,7 @@ func (s *SourceState) newSourceStateFileEntryFromSymlink(actualStateSymlink *Act
 }
 
 // sourceStateEntry returns a new SourceStateEntry based on actualStateEntry.
-func (s *SourceState) sourceStateEntry(actualStateEntry ActualStateEntry, destAbsPath AbsPath, info os.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
+func (s *SourceState) sourceStateEntry(actualStateEntry ActualStateEntry, destAbsPath AbsPath, info fs.FileInfo, parentSourceRelPath SourceRelPath, options *AddOptions) (SourceStateEntry, error) {
 	switch actualStateEntry := actualStateEntry.(type) {
 	case *ActualStateAbsent:
 		return nil, fmt.Errorf("%s: not found", destAbsPath)
